/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: CollectionManager.java,v 1.26 2004/02/21 03:56:11 vgritsenko Exp $
 */

package org.apache.xindice.core;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xindice.util.Configurable;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.util.ConfigurationCallback;
import org.apache.xindice.util.XindiceException;

import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Iterator;

/**
 * CollectionManager is the base class for both Database and Collection.
 *
 * @version CVS $Revision: 1.26 $, $Date: 2004/02/21 03:56:11 $
 */
public class CollectionManager implements Configurable {

    private static final Log log = LogFactory.getLog(CollectionManager.class);

    private static final String COLLECTION = "collection";
    private static final String COLLECTIONS = "collections";
    private static final String[] EMPTY_STRINGS = new String[0];
    private static final String NAME = "name";

    // FIXME: Access to collections and config is not synchronized
    private final Map collections = new HashMap();
    private Configuration config = null;


    protected CollectionManager() {
    }

    /**
     *
     * @param collection
     */
    protected void addCollection(Collection collection) {
        this.collections.put(collection.getName(), collection);
    }

    /**
     * Returns number of child collections
     *
     * @return number of collections
     */
    public final long countCollections() throws DBException {
        return (long) collections.size();
    }

    /**
     * createCollection creates a new Collection object and any associated
     * system resources that the Collection will need.
     *
     * @param path The relative path of the Collection
     * @param cfg The Collection's configuration
     * @return The newly created Collection
     */
    public Collection createCollection(String path, Configuration cfg) throws DBException {
        if (path == null || "".equals(path)) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "Null or empty collection name");
        }

        if (path.indexOf("/") != -1) {
            CollectionManager cm = this;
            StringTokenizer st = new StringTokenizer(path, "/");
            while (cm != null && st.hasMoreTokens()) {
                path = st.nextToken().trim();
                if (path.length() == 0) {
                    continue;
                }

                if (st.hasMoreTokens()) {
                    cm = (CollectionManager) cm.collections.get(path);
                } else {
                    return cm.createCollection(path, cfg);
                }
            }

            throw new DBException(FaultCodes.COL_COLLECTION_NOT_FOUND,
                                  "Parent Collection '" + path + "' doesn't exist");
        }

        if (path.indexOf('/') != -1) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE, "Name cannot contain '/'");
        }

        Collection collection = new Collection((Collection) this);
        try {
            // Do a name check to see if all is well
            String n = cfg.getAttribute(NAME);
            if (n == null || n.trim().equals("")) {
                throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                      "No name specified in collection configuration");
            }
            if (!n.equals(path)) {
                throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                      "Name does not match with path name");
            }

            if (getCollection(n) != null) {
                throw new DBException(FaultCodes.COL_DUPLICATE_COLLECTION,
                                      "Duplicate Collection '" + n + "'");
            }

            Configuration colConfig = this.config.getChild(COLLECTIONS, true);
            colConfig.add(cfg);

            collection.setConfig(cfg);
            collection.create();
            collections.put(n, collection);
            if (log.isInfoEnabled()) {
                log.info("Created a new collection named '" + n + "'");
            }
        } catch (DBException e) {
            // Do not wrap DBException
            throw e;
        } catch (Exception e) {
            throw new DBException(FaultCodes.COL_CANNOT_CREATE,
                                  "Error Creating Collection '" + path + "'", e);
        }
        return collection;
    }

    public boolean close() throws DBException {
        for(Iterator i = collections.values().iterator(); i.hasNext(); ) {
            Collection collection = (Collection)i.next();
            try {
                collection.close();
            } catch (DBException e) {
                if (log.isWarnEnabled()) {
                    log.warn("ignored exception", e);
                }
            }
        }

        return true;
    }

    /**
     * dropCollection physically removes the specified Collection and any
     * associated system resources that the Collection uses.
     *
     * @param collection The Collection to drop
     * @return Whether or not the Collection was dropped
     */
    public boolean dropCollection(Collection collection) throws DBException {
        if (collection == null) {
            throw new DBException(FaultCodes.COL_COLLECTION_NOT_FOUND,
                                  "Collection Value Null");
        }

        Collection cm = collection.getParentCollection();
        if (cm == null) {
            throw new DBException(FaultCodes.DBE_CANNOT_DROP,
                                  "You Cannot Drop The Database");
        }

        if (cm != this) {
            return cm.dropCollection(collection);
        } else {
            final String name = collection.getName();
            boolean dropped = collection.drop();
            if (dropped) {
                collections.remove(name);
                Configuration colConfig = config.getChild(COLLECTIONS);
                colConfig.processChildren(COLLECTION, new ConfigurationCallback() {
                    public void process(Configuration cfg) {
                        try {
                            if (cfg.getAttribute(NAME).equals(name)) {
                                cfg.delete();
                            }
                        } catch (Exception e) {
                            if (log.isWarnEnabled()) {
                                log.warn("ignored exception", e);
                            }
                        }
                    }
                });
            }
            return dropped;
        }
    }

    /**
     * getCollection retrieves a Collection by name.
     *
     * @param path The Collection path
     * @return The Collection (or null)
     */
    public Collection getCollection(String path) throws DBException {
        if (path == null) {
            return null;
        }

        if (path.indexOf("/") != -1) {
            CollectionManager cm = this;
            StringTokenizer st = new StringTokenizer(path, "/");
            while (cm != null && st.hasMoreTokens()) {
                path = st.nextToken();
                cm = (CollectionManager) cm.collections.get(path);
            }
            return (Collection) cm;
        } else {
            return (Collection) collections.get(path);
        }
    }

    /**
     * @return The collections map
     */
    protected Map getCollections() {
        return this.collections;
    }

    /**
     * @see org.apache.xindice.util.Configurable#getConfig()
     */
    public Configuration getConfig() {
        return config;
    }

    /**
     * listCollections retrieves a list of Collections as an array of
     * Strings.
     *
     * @return The Collection list
     */
    public final String[] listCollections() throws DBException {
        return (String[]) collections.keySet().toArray(EMPTY_STRINGS);
    }

    /**
     * @see org.apache.xindice.util.Configurable#setConfig(org.apache.xindice.util.Configuration)
     */
    public void setConfig(Configuration config) throws XindiceException {
        this.config = config;

        Configuration colConfig = config.getChild(COLLECTIONS);
        if (colConfig != null) {
            colConfig.processChildren(COLLECTION, new ConfigurationCallback() {
                public void process(Configuration cfg) throws XindiceException {
                    // check for an existing Collection by name and, if found, skip creating a new Collection
                    // else, create a new child Collection, configure it an add it
                    // if the Collection already exists in our collections list,
                    // creating a new one will cause the old one to be discarded
                    // while holding open file handles, etc.
                    Collection col = (Collection) collections.get(cfg.getAttribute(NAME));
                    if (col == null) {
                        col = new Collection((Collection) CollectionManager.this);
                        col.setConfig(cfg);
                        collections.put(col.getName(), col);
                    }
                    // else, assume col is configured elsewhere...
                    // I'm not sure this should be happening, but it does
                    // it is not safe to call col.setConfig again since it does a bunch
                    // of stuff (like creating a new Filer and overwriting the existing one, etc.)
                }
            });
        }
    }
}
