/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: ImportTree.java,v 1.12 2004/02/08 02:57:35 vgritsenko Exp $
 */

package org.apache.xindice.tools.command;

import org.apache.xindice.tools.XMLTools;

import org.xmldb.api.DatabaseManager;
import org.xmldb.api.base.Collection;

import java.io.File;
import java.io.FileFilter;
import java.util.Hashtable;

/**
 * ImportTree.java is designed to take a directory/file path and create a Collection
 * from it's name. The files within the directory are converted into Documents
 * and then stored into the newly created Collection with their filename as a document key
 *
 * NOTE:  If the collection passed in with the -c option already exists it will be overwritten!
 *
 * @version CVS $Revision: 1.12 $, $Date: 2004/02/08 02:57:35 $
 */
public class ImportTree extends Command {

    Command addDocument = null;
    Command addCollection = null;

    public boolean execute(Hashtable table) throws Exception {

        Collection col = null;
        try {
            // Setup the creation commands we'll use to add documents and create
            // collections.
            addDocument =
                    (Command) Class.forName("org.apache.xindice.tools.command.AddDocument").newInstance();

            addCollection =
                    (Command) Class.forName("org.apache.xindice.tools.command.AddCollection").newInstance();

            String startcollection = null;

            // Verify that the correct command line parameters were passed in
            if (table.get(XMLTools.FILE_PATH) == null) {
                System.out.println("ERROR : Directory name and switch required");
                return false;
            }
            if (table.get(XMLTools.COLLECTION) == null) {
                System.out.println("ERROR : Collection name and switch required");
                return false;
            }

            startcollection = (String) table.get(XMLTools.COLLECTION);

            // Make sure the root collection exists
            String colstring = normalizeCollectionURI(startcollection,
                                                      (String) table.get(XMLTools.LOCAL));
            if ((col = DatabaseManager.getCollection(colstring)) == null) {
                System.out.println("ERROR : Collection not found!");
                return false;
            }

            File startdir = new File((String) table.get(XMLTools.FILE_PATH));

            // Make call to process, this is called recursively!
            process(startdir, colstring, table);

        } finally {
            // Be sure to close collection objects
            if (col != null) {
                col.close();
            }
        }

        return true;
    }

    // Recursive function to create collections/documents
    void process(File directory, String baseCollection, Hashtable table) {
        try {
            // The file extension to use for reading in files
            final String ext = (String) table.get(XMLTools.EXTENSION);

            // Make sure we received a directory so we don't create a collection
            // named after a file.
            if (!directory.isDirectory()) {
                System.out.println("ERROR: The specified import path is not a directory " +
                                   directory.getPath());
                return;
            }

            // Create the collection for this directory.
            baseCollection = createCollection(baseCollection, directory.getName());

            // Get a directory listing and loop over
            File[] subdirs = directory.listFiles(new ExtensionFileFilter(ext));

            for (int i = 0; i < subdirs.length; i++) {
                // Is this a directory?  If so, call ourselves recursively
                if (subdirs[i].isDirectory()) {
                    process(subdirs[i], baseCollection, table);
                } else {
                    // Otherwise we have a file so import it.
                    importFile(baseCollection, subdirs[i].getAbsolutePath(),
                               subdirs[i].getName());
                }
            }
        } catch (Exception e) {
            System.out.println("ERROR : " + e.getMessage());
            if (table.get(XMLTools.VERBOSE).equals("true")) {
                e.printStackTrace(System.err);
            }
            return;
        }
    }

    protected void importFile(String baseCollection, String path, String name)
            throws Exception {
        Hashtable table = new Hashtable();
        // Use functionality from AddDocument to add this document
        table.put(XMLTools.COLLECTION, baseCollection);
        table.put(XMLTools.FILE_PATH, path);
        table.put(XMLTools.NAME_OF, name);

        addDocument.execute(table);
    }

    protected String createCollection(String baseCollection, String newCollection)
            throws Exception {
        if (!newCollection.equals(".")) {
            Hashtable table = new Hashtable();
            table.put(XMLTools.COLLECTION, baseCollection);
            table.put(XMLTools.NAME_OF, newCollection);

            addCollection.execute(table);
        } else {
            newCollection = "";
        }

        return baseCollection + "/" + newCollection;
    }

    /**
     * Allow us to filter directory listings over certain file extensions
     */
    private static class ExtensionFileFilter implements FileFilter {

        private String extension = "";

        /**
         * Constructs an <code>MyFileFiler</code> instance with a specific
         * extension.
         *
         * @param extension The file's extension
         */
        public ExtensionFileFilter(String extension) {
            this.extension = extension;
        }

        /**
         * Tests whether or not the specified abstract pathname should be
         * included in a pathname list.
         *
         * @param  pathname  The abstract pathname to be tested
         * @return  <code>true</code> if and only if <code>pathname</code>
         *          should be included
         */
        public boolean accept(File pathname) {
            if (this.extension != "") {
                return pathname.getName().endsWith("." + this.extension) || pathname.isDirectory();
            } else {
                return true;
            }
        }
    }

}
