/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: XMLCompressedInput.java,v 1.11 2004/02/08 03:50:13 vgritsenko Exp $
 */

package org.apache.xindice.xml;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Node;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;

/**
 * XMLCompressedInput is an InputStream extension that provides functions for
 * reading type from a Xindice Compressed XML Stream.
 *
 * @version CVS $Revision: 1.11 $, $Date: 2004/02/08 03:50:13 $
 */
public class XMLCompressedInput extends DataInputStream {

    private static final Log log = LogFactory.getLog(XMLCompressedInput.class);

    private byte signature = 0;
    // Not used: private SymbolTable st = null;

    public XMLCompressedInput(InputStream is, SymbolTable st) {
        super(is);
        // this.st = st;
    }

    /**
     * readSignature reads a Node signature byte from the input stream.
     *
     * @return The signature
     * @throws IOException if an IOException occurs
     */
    public final byte readSignature() throws IOException {
        signature = readByte();
        return signature;
    }

    /**
     * getSignature returns the most recently read signature byte from the
     * input stream.
     *
     * @return The signature
     */
    public final byte getSignature() {
        return signature;
    }

    /**
     * getNodeType resolves and returns the DOM Node type from the given
     * Xindice compressed signature.
     *
     * @return The DOM Node type
     */
    public final short getNodeType() {
        byte type = (byte) ((signature & 0xC0) >>> 0x6);
        switch (type) {

            case Signatures.Char:
                type = (byte) ((signature & 0x20) >>> 0x5);
                if (type == Signatures.CharText) {
                    return Node.TEXT_NODE;
                } else {
                    return Node.ENTITY_REFERENCE_NODE;
                }
                
            case Signatures.Elem:
                return Node.ELEMENT_NODE;

            case Signatures.Proc:
                return Node.PROCESSING_INSTRUCTION_NODE;

            case Signatures.Decl:
                type = (byte) ((signature & 0x3C) >>> 0x2);
                switch (type) {

                    case Signatures.DeclCData:
                        return Node.CDATA_SECTION_NODE;

                    case Signatures.DeclEntity:
                        return Node.ENTITY_NODE;

                    case Signatures.DeclComment:
                        return Node.COMMENT_NODE;

                    case Signatures.DeclDocType:
                        return Node.DOCUMENT_TYPE_NODE;

                    case Signatures.DeclNotation:
                        return Node.NOTATION_NODE;

                    default:
                        return Signatures.Unknown;

                }
            default:
                if (log.isWarnEnabled()) {
                    log.warn("invalid signature : " + type);
                }

        }
        return Signatures.Unknown;
    }

    /**
     * readContentSize reads the size of the upcoming content from the Stream,
     * based on the object signature that is provided.
     *
     * @return The Content Size
     * @throws IOException if an IOException occurs
     */
    public final int readContentSize() throws IOException {
        int sizeType = 0;
        short nodeType = getNodeType();
        switch (nodeType) {

            case Node.TEXT_NODE:
            case Node.ELEMENT_NODE:
                sizeType = (byte) ((signature & 0x0C) >>> 0x2);
                break;

            case Node.PROCESSING_INSTRUCTION_NODE:
            case Node.COMMENT_NODE:
            case Node.CDATA_SECTION_NODE:
                sizeType = Signatures.IntContent;
                break;

            case Node.NOTATION_NODE:
                sizeType = (byte) (signature & 0x03);
                break;
            default:
                if (log.isWarnEnabled()) {
                    log.warn("invalid node type : " + nodeType);
                }
        }

        switch (sizeType) {

            case Signatures.ByteContent:
                return readByte();

            case Signatures.ShortContent:
                return readShort();

            case Signatures.IntContent:
                return readInt();

            case Signatures.NoContent:
                // Do Nothing
				break;
				
            default:
                if (log.isWarnEnabled()) {
                    log.warn("invalid size : " + sizeType);
                }
        }
        return 0;
    }

    /**
     * readAttributeCount reads the count of the upcoming Attributes from the
     * Stream based on the object signature that is provided.
     *
     * @return The Content Size
     * @throws IOException if an IOException occurs
     */
    public final int readAttributeCount() throws IOException {
        int countType = 0;
        if (getNodeType() == Node.ELEMENT_NODE) {
            countType = (byte) (signature & 0x03);
        }

        switch (countType) {

            case Signatures.ByteContent:
                return readByte();

            case Signatures.ShortContent:
                return readShort();

            case Signatures.IntContent:
                return readInt();

            case Signatures.NoContent:
                // Do Nothing
				break;
				
            default:
                if (log.isWarnEnabled()) {
                    log.warn("invalid signature type : " + countType);
                }
        }
        return 0;
    }
}
