/*
 * Copyright 1999-2004 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * CVS $Id: FilerTestBase.java,v 1.3 2004/02/08 03:56:47 vgritsenko Exp $
 */

package org.apache.xindice.core.filer;

import org.apache.xindice.core.data.Key;
import org.apache.xindice.core.data.Record;
import org.apache.xindice.core.data.RecordSet;
import org.apache.xindice.core.data.Value;
import org.apache.xindice.util.Configuration;
import org.apache.xindice.xml.dom.DOMParser;

import junit.framework.TestCase;

import java.util.List;
import java.util.Vector;
import java.io.File;

/**
 * Base class for filer test cases
 *
 * @version CVS $Revision: 1.3 $, $Date: 2004/02/08 03:56:47 $
 * @author <a href="mailto:kstaken@xmldatabases.org">Kimbro Staken</a>
 * @author <a href="mailto:vladimir@apache.org">Vladimir R. Bossicard</a>
 * @author <a href="mailto:vgritsenko@apache.org">Vadim Gritsenko</a>
 */
public class FilerTestBase extends TestCase {

    static public final String TEST_COLLECTION_NAME = "tests";
    static public final Key TEST_KEY = new Key("test.xml");
    static public final Value TEST_VALUE = new Value("<test><test></test><test>Just a test</test></test>");
    static public final Value TEST_VALUE_2 = new Value("<test><test></test><test>Just a test 2</test></test>");
    static public final Value TEST_VALUE_3 = new Value("<test><test></test><test>Just a test 3</test></test>");

    /**
     * Key size for the testLargeKey. FSFiler will set smaller value for its test.
     */
    protected int LARGE_KEY_SIZE = 8192 + 32;


    private static final File root = new File(TEST_COLLECTION_NAME);
    private Filer filer;

    public FilerTestBase(String name, Filer filer) {
        super(name);
        this.filer = filer;
    }

    public void setUp() throws Exception {
        root.mkdir();
        filer.setLocation(root, TEST_COLLECTION_NAME);
        // FSFiler needs location attribute
        filer.setConfig(new Configuration(DOMParser.toDocument("<filer location=\"" + TEST_COLLECTION_NAME + "\" />")));
        if (!filer.exists()) {
            filer.create();
        }
        filer.open();

        RecordSet set = filer.getRecordSet();
        while (set.hasMoreRecords()) {
            Key key = set.getNextKey();
            filer.deleteRecord(key);
        }
        assertEquals(0, filer.getRecordCount());
    }

    protected void tearDown() throws Exception {
        if (filer != null) {
            filer.close();
        }

        String[] files = root.list();
        for (int i = 0; i < files.length; i++) {
            new File(root, files[i]).delete();
        }
        root.delete();
        super.tearDown();
    }

    public String getName() {
        return this.filer.getName() + " " + super.getName();
    }

    public void testSuccessWriteRecord() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        Record result = filer.readRecord(TEST_KEY);
        assertEquals(TEST_VALUE, result.getValue());

        RecordSet set = filer.getRecordSet();
        assertEquals(result.getValue(), set.getNextRecord().getValue());
        assertTrue(set.hasMoreRecords() == false);

        filer.deleteRecord(TEST_KEY);
        assertEquals(0, filer.getRecordCount());
    }

    public void testFailReadDeletedRecord() throws Exception {
        Record result = filer.readRecord(TEST_KEY);
        assertNull(result);
    }

    public void testFailWriteRecordNullKey() throws Exception {
        try {
            filer.writeRecord(null, TEST_VALUE);
            fail("Expecting FilerException");
        } catch (FilerException e) {
            // Ok
        }
    }

    public void testFailWriteRecordNullValue() throws Exception {
        try {
            filer.writeRecord(TEST_KEY, null);
            fail("Expecting FilerException");
        } catch (FilerException e) {
            // Ok
        }
    }

    public void testFailWriteRecordNullValueKey() throws Exception {
        try {
            filer.writeRecord(null, null);
            fail("Expecting FilerException");
        } catch (FilerException e) {
            // Ok
        }
    }

    public void testReadRecord() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // Valid key
        Record result = filer.readRecord(TEST_KEY);
        assertEquals(TEST_VALUE, result.getValue());
    }

    public void testFailReadRecordNullKey() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // Null key
        Record result = filer.readRecord(null);
        assertNull(result);
    }

    public void testFailReadRecordEmptyKey() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // Empty key
        Record result = filer.readRecord(new Key(""));
        assertNull(result);
    }

    public void testFailReadRecordUnknownKey() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // Non-existant key
        Record result = filer.readRecord(new Key("non-existant-key"));
        assertNull(result);
    }

    public void testDeleteRecord() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // Valid key
        filer.deleteRecord(TEST_KEY);
        assertEquals(0, filer.getRecordCount());
        Record result = filer.readRecord(TEST_KEY);
        assertNull(result);
    }

    public void testDeleteRecordNullKey() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // These should all just fail silently.
        assertTrue(false == filer.deleteRecord(null));
        assertEquals(1, filer.getRecordCount());
    }

    public void testDeleteRecordEmptyKey() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // These should all just fail silently.
        assertTrue(false == filer.deleteRecord(new Key("")));
        assertEquals(1, filer.getRecordCount());
    }

    public void testDeleteRecordWrongKey() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertEquals(1, filer.getRecordCount());

        // These should all just fail silently.
        assertTrue(false == filer.deleteRecord(new Key("non-existant-key")));
        assertEquals(1, filer.getRecordCount());
    }

    public void testGetRecordCount() throws Exception {
        assertEquals(0, filer.getRecordCount());

        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertTrue(filer.writeRecord(new Key("test1"), TEST_VALUE));
        assertTrue(filer.writeRecord(new Key("test2"), TEST_VALUE));
        assertEquals(3, filer.getRecordCount());

        assertTrue(filer.writeRecord(new Key("test3"), TEST_VALUE));
        assertEquals(4, filer.getRecordCount());

        assertTrue(filer.deleteRecord(new Key("test3")));
        assertEquals(3, filer.getRecordCount());

        assertTrue(filer.deleteRecord(TEST_KEY));
        assertTrue(filer.deleteRecord(new Key("test1")));
        assertTrue(filer.deleteRecord(new Key("test2")));

        assertEquals(0, filer.getRecordCount());
    }

    public void testGetRecordSet() throws Exception {
        assertTrue(filer.writeRecord(TEST_KEY, TEST_VALUE));
        assertTrue(filer.writeRecord(new Key("test2"), TEST_VALUE_2));
        assertTrue(filer.writeRecord(new Key("test3"), TEST_VALUE_3));

        RecordSet result = filer.getRecordSet();
        assertNotNull(result);

        List results = new Vector();
        while (result.hasMoreRecords()) {
            results.add(result.getNextRecord().getValue());
        }
        assertEquals(3, results.size());
        assertTrue(results.contains(TEST_VALUE));
        assertTrue(results.contains(TEST_VALUE_2));
        assertTrue(results.contains(TEST_VALUE_3));

        assertTrue(filer.deleteRecord(TEST_KEY));
        assertTrue(filer.deleteRecord(new Key("test2")));
        assertTrue(filer.deleteRecord(new Key("test3")));

        result = filer.getRecordSet();
        assertTrue(!result.hasMoreRecords());
    }

    public void testInsertManyDocuments() throws Exception {
        int iterations = 1000;
        for (int i = 0; i < iterations; i++) {
            assertTrue(filer.writeRecord(new Key("key" + i), TEST_VALUE));
        }

        assertTrue(filer.getRecordCount() == iterations);

        for (int i = 0; i < iterations; i++) {
            assertTrue(filer.deleteRecord(new Key("key" + i)));
        }

        assertTrue(filer.getRecordCount() == 0);
    }

    public void testConcurrentInsert() throws Exception {
        assertTrue(filer.getRecordCount() == 0);

        final int THREADS = 30;
        final int ITERATIONS = 65;

        Thread[] threads = new Thread[THREADS];
        for (int i = 0; i < THREADS; i++) {
            final int threadID = i;
            threads[i] = new Thread() {
                public void run() {
                    for (int ii = 0; ii < ITERATIONS; ii++) {
                        Key key = new Key("T" + threadID + "I" + ii);
                        Value value = new Value("<test thread=\"" + threadID + "\" iteration=\"" + ii + "\"/>");
                        try {
                            filer.writeRecord(key, value);
                        } catch (Exception e) {
                            e.printStackTrace();
                        }
                    }
                    // System.out.println(getName() + " done.");
                }
            };
            threads[i].setName("FilerTest" + i);
        }

        // Start all the threads at once
        for (int i = 0; i < THREADS; i++) {
            threads[i].start();
        }
        Thread.sleep(1000);

        for (int i = 0; i < THREADS; i++) {
            threads[i].join();
        }

        filer.flush();

        // Check results
        assertEquals(filer.getRecordCount(), THREADS * ITERATIONS);
        for (int i = 0; i < THREADS; i++) {
            for (int ii = 0; ii < ITERATIONS; ii++) {
                Key key = new Key("T" + i + "I" + ii);
                Value value = new Value("<test thread=\"" + i + "\" iteration=\"" + ii + "\"/>");
                Record record = filer.readRecord(key);
                assertNotNull("Record with key '" + key + "' was not found",
                              record);
                assertEquals("Expected record with key '" + key + "', found record with key '" + record.getKey() + "'",
                             key, record.getKey());
                assertEquals("Expected record with value '" + value + "', found record with value '" + record.getValue() + "'",
                             value, record.getValue());
            }
        }
    }

    public void testLargeKey() throws Exception {
        assertTrue(filer.getRecordCount() == 0);

        StringBuffer sb = new StringBuffer(LARGE_KEY_SIZE);
        sb.append("KEY");
        for (int k = 0; k < LARGE_KEY_SIZE - 3; k++) {
            sb.append('0');
        }
        final Key key = new Key(sb.toString());
        final Value value = new Value("<test/>");

        assertTrue(filer.writeRecord(key, value));

        assertEquals(filer.getRecordCount(), 1);
        Record record = filer.readRecord(key);
        assertNotNull("Record with key '" + key + "' was not found",
                      record);
        assertEquals("Expected record with key '" + key + "', found record with key '" + record.getKey() + "'",
                     key, record.getKey());
        assertEquals("Expected record with value '" + value + "', found record with value '" + record.getValue() + "'",
                     value, record.getValue());
    }
}
